@@ -49,7 +49,6 @@
 
 typedef struct _Cache
 {
-    int n_ref;
     char md5[33]; /* cache id */
     /* environment */
     char* menu_name;
@@ -63,49 +62,60 @@ typedef struct _Cache
     /* GFileMonitor* cache_mon; */
 
     guint delayed_reload_handler;
+    guint delayed_free_handler;
 
     /* clients requesting this menu cache */
     GSList* clients;
 }Cache;
 
+GMainLoop* main_loop = NULL;
+
 static GHashTable* hash = NULL;
 
 static void on_file_changed( GFileMonitor* mon, GFile* gf, GFile* other,
                              GFileMonitorEvent evt, Cache* cache );
 
-static void cache_unref(Cache* cache)
+static gboolean delayed_cache_free(gpointer data)
 {
-    --cache->n_ref;
-    /* DEBUG("menu cache unrefed: ref_count=%d", cache->n_ref); */
-    if( cache->n_ref == 0 )
+    Cache* cache = data;
+    int i;
+
+    if(g_source_is_destroyed(g_main_current_source()))
+        return FALSE;
+
+    g_hash_table_remove( hash, cache->md5 );
+    /* DEBUG("menu cache freed"); */
+    for(i = 0; i < cache->n_files; ++i)
     {
-        int i;
-        /* DEBUG("menu cache freed"); */
-        g_slist_foreach( cache->clients, (GFunc)g_io_channel_unref, NULL );
-        g_slist_free( cache->clients );
-        for(i = 0; i < cache->n_files; ++i)
-        {
-            g_file_monitor_cancel( cache->mons[i] );
-            g_object_unref( cache->mons[i] );
-        }
+        g_file_monitor_cancel( cache->mons[i] );
+        g_object_unref( cache->mons[i] );
+    }
 /*
-        g_file_monitor_cancel(cache->cache_mon);
-        g_object_unref(cache->cache_mon);
+    g_file_monitor_cancel(cache->cache_mon);
+    g_object_unref(cache->cache_mon);
 */
-        g_free( cache->mons );
-        g_strfreev( cache->env );
-        g_strfreev( cache->files );
+    g_free( cache->mons );
+    g_strfreev( cache->env );
+    g_strfreev( cache->files );
 
-		if( cache->delayed_reload_handler )
-			g_source_remove( cache->delayed_reload_handler );
+    if( cache->delayed_reload_handler )
+        g_source_remove( cache->delayed_reload_handler );
 
-        g_slice_free( Cache, cache );
-    }
-    else if( cache->n_ref == 1 ) /* the last ref count is held by hash table */
-    {
-        /* DEBUG("menu cache removed from hash"); */
-        g_hash_table_remove( hash, cache->md5 );
-    }
+    g_slice_free( Cache, cache );
+
+    if(g_hash_table_size(hash) == 0)
+        g_main_loop_quit(main_loop);
+    return FALSE;
+}
+
+static void cache_free(Cache* cache)
+{
+    /* shutdown cache in 10 minutes of inactivity */
+    if(!cache->delayed_free_handler)
+        cache->delayed_free_handler = g_timeout_add_seconds(600,
+                                                            delayed_cache_free,
+                                                            cache);
+    DEBUG("menu %p cache unused, removing in 600s", cache);
 }
 
 static gboolean read_all_used_files( FILE* f, int* n_files, char*** used_files )
@@ -159,7 +169,7 @@ static void set_env( char* penv, const char* name )
 
 static void pre_exec( gpointer user_data )
 {
-    char** env = (char*)user_data;
+    char** env = (char**)user_data;
     set_env(*env, "XDG_CACHE_HOME");
     set_env(*++env, "XDG_CONFIG_DIRS");
     set_env(*++env, "XDG_MENU_PREFIX");
@@ -178,7 +188,7 @@ static gboolean regenerate_cache( const char* menu_name,
     FILE* f;
     int n_files, status = 0;
     char** files;
-    const char* const * argv[] = {
+    const char* argv[] = {
         MENUCACHE_LIBEXECDIR "/menu-cache-gen",
         "-l", NULL,
         "-i", NULL,
@@ -191,7 +201,8 @@ static gboolean regenerate_cache( const char* menu_name,
     /* DEBUG("cmd: %s", g_strjoinv(" ", argv)); */
 
     /* run menu-cache-gen */
-    if( !g_spawn_sync(NULL, argv, NULL, 0,
+    /* FIXME: is cast to (char**) valid here? */
+    if( !g_spawn_sync(NULL, (char **)argv, NULL, 0,
                     pre_exec, env, NULL, NULL, &status, NULL ))
     {
         DEBUG("error executing menu-cache-gen");
@@ -226,6 +237,9 @@ static gboolean delayed_reload( Cache* cache )
     int i;
     GFile* gf;
 
+    if(g_source_is_destroyed(g_main_current_source()))
+        return FALSE;
+
     /* DEBUG("Re-generation of cache is needed!"); */
     /* DEBUG("call menu-cache-gen to re-generate the cache"); */
     memcpy( buf, "REL:", 4 );
@@ -262,7 +276,8 @@ static gboolean delayed_reload( Cache* cache )
             cache->mons[i] = g_file_monitor_directory( gf, 0, NULL, NULL );
         else
             cache->mons[i] = g_file_monitor_file( gf, 0, NULL, NULL );
-        g_signal_connect( cache->mons[i], "changed", on_file_changed, cache);
+        g_signal_connect(cache->mons[i], "changed",
+                         G_CALLBACK(on_file_changed), cache);
         g_object_unref(gf);
     }
 /*
@@ -277,7 +292,8 @@ static gboolean delayed_reload( Cache* cache )
     for( l = cache->clients; l; l = l->next )
     {
         GIOChannel* ch = (GIOChannel*)l->data;
-        write(g_io_channel_unix_get_fd(ch), buf, 37 );
+        if(write(g_io_channel_unix_get_fd(ch), buf, 37) < 37)
+            g_io_channel_shutdown(ch, FALSE, NULL);
     }
     cache->delayed_reload_handler = 0;
     return FALSE;
@@ -402,7 +418,6 @@ void on_file_changed( GFileMonitor* mon, GFile* gf, GFile* other,
                     /* FIXME: temporarily disable monitor of the cached file before utime() */
                     /* without this, directory mtime will > mtime of cache file,
                      * and the menu will get re-generated unnecessarily the next time. */
-                    GFile* gf;
                     struct utimbuf ut;
                     char* cache_file;
 
@@ -412,22 +427,22 @@ void on_file_changed( GFileMonitor* mon, GFile* gf, GFile* other,
 
                     ut.actime = ut.modtime = time(NULL);
                     /* stop monitor of cache file. */
-					/*
+                    /*
                     g_file_monitor_cancel( cache->cache_mon );
                     g_object_unref( cache->cache_mon );
                     cache->cache_mon = NULL;
-					*/
+                    */
                     /* update the mtime of the cache file. */
                     utime( cache_file, &ut );
 
                     /* restart the monitor */
                     /* gf = g_file_new_for_path( cache_file ); */
                     g_free(cache_file);
-					/*
+                    /*
                     cache->cache_mon = g_file_monitor_file(gf, 0, NULL, NULL);
                     g_object_unref(gf);
                     g_signal_connect( cache->cache_mon, "changed", G_CALLBACK(on_file_changed), cache);
-					*/
+                    */
                     DEBUG("files are changed, but no re-generation is needed.");
                     return;
                 }
@@ -442,15 +457,14 @@ void on_file_changed( GFileMonitor* mon, GFile* gf, GFile* other,
     cache->delayed_reload_handler = g_timeout_add_seconds_full( G_PRIORITY_LOW, 3, (GSourceFunc)delayed_reload, cache, NULL );
 }
 
-static gboolean cache_file_is_updated( const char* cache_file, int* n_used_files, char** used_files )
+static gboolean cache_file_is_updated( const char* cache_file, int* n_used_files, char*** used_files )
 {
     gboolean ret = FALSE;
     struct stat st;
     time_t cache_mtime;
     char** files;
-    int n, i, l;
+    int n, i;
     FILE* f;
-    char line[ 4096 ];
 
     f = fopen( cache_file, "r" );
     if( f )
@@ -476,7 +490,6 @@ static gboolean cache_file_is_updated( const char* cache_file, int* n_used_files
                 }
             }
         }
-_out:
         fclose( f );
     }
     return ret;
@@ -493,7 +506,8 @@ static void get_socket_name( char* buf, int len )
         if(*p)
             *p = '\0';
     }
-    g_snprintf( buf, len, "/tmp/.menu-cached-%s-%s", dpy ? dpy : ":0", g_get_user_name() );
+    g_snprintf( buf, len, "%s/.menu-cached-%s-%s", g_get_tmp_dir(),
+                dpy ? dpy : ":0", g_get_user_name() );
     g_free(dpy);
 }
 
@@ -505,7 +519,7 @@ static int create_socket()
     fd = socket(PF_UNIX, SOCK_STREAM, 0);
     if (fd < 0)
     {
-        g_print("Failed to create socket\n");
+        DEBUG("Failed to create socket");
         return -1;
     }
     memset(&addr, 0, sizeof(addr));
@@ -514,12 +528,12 @@ static int create_socket()
 
     /* remove previous socket file */
     if (unlink(addr.sun_path) < 0) {
-	if (errno != ENOENT)
-	    g_error("Couldn't remove previous socket file %s", addr.sun_path);
+        if (errno != ENOENT)
+            g_error("Couldn't remove previous socket file %s", addr.sun_path);
     }
     /* remove of previous socket file successful */
     else
-	    g_warning("removed previous socket file %s", addr.sun_path);
+        g_warning("removed previous socket file %s", addr.sun_path);
 
     if(bind(fd, (struct sockaddr *) &addr, sizeof(addr)) < 0)
     {
@@ -536,17 +550,18 @@ static int create_socket()
     return fd;
 }
 
-static void on_client_closed(GIOChannel* ch, gpointer user_data)
+static void on_client_closed(gpointer user_data)
 {
+    GIOChannel* ch = user_data;
     GHashTableIter it;
     char* md5;
     Cache* cache;
-    GSList *l, *to_remove = NULL;
+    GSList *l;
     DEBUG("client closed: %p", ch);
     g_hash_table_iter_init (&it, hash);
     while( g_hash_table_iter_next (&it, (gpointer*)&md5, (gpointer*)&cache) )
     {
-        if( l = g_slist_find( cache->clients, ch ) )
+        while((l = g_slist_find( cache->clients, ch )) != NULL)
         {
             /* FIXME: some clients are closed accidentally without
              * unregister the menu first due to crashes.
@@ -557,15 +572,14 @@ static void on_client_closed(GIOChannel* ch, gpointer user_data)
              * unref will be called twice and incorrect ref. counting
              * will happen.
              */
-            to_remove = g_slist_prepend( to_remove, cache );
             cache->clients = g_slist_delete_link( cache->clients, l );
-            DEBUG("remove channel from cache");
-            g_io_channel_unref(ch);
+            DEBUG("remove channel from cache %p", cache);
+            if(cache->clients == NULL)
+                cache_free(cache);
         }
     }
     /* DEBUG("client closed"); */
-    g_slist_foreach( to_remove, (GFunc)cache_unref, NULL );
-    g_slist_free( to_remove );
+    g_io_channel_unref(ch);
 }
 
 static gboolean on_client_data_in(GIOChannel* ch, GIOCondition cond, gpointer user_data)
@@ -576,10 +590,10 @@ static gboolean on_client_data_in(GIOChannel* ch, GIOCondition cond, gpointer us
     const char* md5;
     Cache* cache;
     GFile* gf;
+    gboolean ret = TRUE;
 
     if(cond & (G_IO_HUP|G_IO_ERR) )
     {
-        on_client_closed(ch, user_data);
         return FALSE; /* remove the watch */
     }
 
@@ -597,7 +611,7 @@ retry:
 
     if( memcmp(line, "REG:", 4) == 0 )
     {
-        int status = 0, n_files, i;
+        int n_files, i;
         char *pline = line + 4;
         char *sep, *menu_name, *lang_name, *cache_dir, *cache_file;
         char **files;
@@ -632,7 +646,7 @@ retry:
 
             env = g_strsplit(pline, "\t", 0);
 
-			cache_dir = env[0]; /* XDG_CACHE_HOME */
+            cache_dir = env[0]; /* XDG_CACHE_HOME */
             /* obtain cache dir from client's env */
 
             cache_file = g_build_filename(*cache_dir ? cache_dir : g_get_user_cache_dir(), "menus", md5, NULL );
@@ -662,24 +676,29 @@ retry:
                 else
                     cache->mons[i] = g_file_monitor_file( gf, 0, NULL, NULL );
                 DEBUG("monitor: %s", g_file_get_path(gf));
-                g_signal_connect( cache->mons[i], "changed", on_file_changed, cache);
+                g_signal_connect(cache->mons[i], "changed",
+                                 G_CALLBACK(on_file_changed), cache);
                 g_object_unref(gf);
             }
-			/*
+            /*
             gf = g_file_new_for_path( cache_file );
             cache->cache_mon = g_file_monitor_file( gf, 0, NULL, NULL );
             g_signal_connect( cache->cache_mon, "changed", on_file_changed, cache);
             g_object_unref(gf);
-			*/
+            */
             g_free(cache_file);
 
             g_hash_table_insert(hash, cache->md5, cache);
-            DEBUG("new menu cache added to hash");
-            cache->n_ref = 1;
+            DEBUG("new menu cache %p added to hash", cache);
         }
         /* DEBUG("menu %s requested by client %d", md5, g_io_channel_unix_get_fd(ch)); */
-        ++cache->n_ref;
-        cache->clients = g_slist_prepend( cache->clients, g_io_channel_ref(ch) );
+        cache->clients = g_slist_prepend(cache->clients, ch);
+        if(cache->delayed_free_handler)
+        {
+            g_source_remove(cache->delayed_free_handler);
+            cache->delayed_free_handler = 0;
+        }
+        DEBUG("client %p added to cache %p", ch, cache);
 
         /* generate a fake reload notification */
         DEBUG("fake reload!");
@@ -689,26 +708,26 @@ retry:
         reload_cmd[37] = '\0';
 
         DEBUG("reload command: %s", reload_cmd);
-        write(g_io_channel_unix_get_fd(ch), reload_cmd, 37);
+        ret = write(g_io_channel_unix_get_fd(ch), reload_cmd, 37) > 0;
     }
     else if( memcmp(line, "UNR:", 4) == 0 )
     {
         md5 = line + 4;
         DEBUG("unregister: %s", md5);
         cache = (Cache*)g_hash_table_lookup(hash, md5);
-        if( cache )
+        if(cache && cache->clients)
         {
             /* remove the IO channel from the cache */
             cache->clients = g_slist_remove(cache->clients, ch);
-            cache_unref(cache);
-            g_io_channel_unref(ch);
+            if(cache->clients == NULL)
+                cache_free(cache);
         }
         else
             DEBUG("bug! client is not found.");
     }
     g_free( line );
 
-    return TRUE;
+    return ret;
 }
 
 static gboolean on_new_conn_incoming(GIOChannel* ch, GIOCondition cond, gpointer user_data)
@@ -729,10 +748,12 @@ static gboolean on_new_conn_incoming(GIOChannel* ch, GIOCondition cond, gpointer
     }
 
     child = g_io_channel_unix_new(client);
-    g_io_add_watch( child, G_IO_PRI|G_IO_IN|G_IO_HUP|G_IO_ERR, on_client_data_in, NULL );
     g_io_channel_set_close_on_unref( child, TRUE );
+    g_io_add_watch_full(child, G_PRIORITY_DEFAULT, G_IO_PRI|G_IO_IN|G_IO_HUP|G_IO_ERR,
+                        on_client_data_in, child, on_client_closed);
+    g_io_channel_unref(child);
 
-    /* DEBUG("new client accepted"); */
+    DEBUG("new client accepted: %p", child);
     return TRUE;
 }
 
@@ -750,6 +771,7 @@ static gboolean on_server_conn_close(GIOChannel* ch, GIOCondition cond, gpointer
 {
     /* FIXME: is this possible? */
     /* the server socket is accidentally closed. terminate the server. */
+    g_io_channel_unref(ch);
     terminate(SIGTERM);
     return TRUE;
 }
@@ -757,12 +779,14 @@ static gboolean on_server_conn_close(GIOChannel* ch, GIOCondition cond, gpointer
 
 int main(int argc, char** argv)
 {
-    GMainLoop* main_loop = g_main_loop_new( NULL, TRUE );
     GIOChannel* ch;
-    int fd, pid, server_fd;
+    int server_fd;
+#ifndef DISABLE_DAEMONIZE
+    int fd, pid;
 
     long open_max;
     long i;
+#endif
 
     server_fd = create_socket();
 
@@ -772,24 +796,24 @@ int main(int argc, char** argv)
 #ifndef DISABLE_DAEMONIZE
     /* Become a daemon */
     if ((pid = fork()) < 0) {
-    	g_error("can't fork");
+        g_error("can't fork");
     }
     else if (pid != 0) {
-	/* exit parent */
-    	exit(0);
+        /* exit parent */
+        exit(0);
     }
 
     /* change working directory to root, so previous working directory
      * can be unmounted */
     if (chdir("/") < 0) {
-    	g_error("can't change directory to /");
+        g_error("can't change directory to /");
     }
 
     open_max = sysconf (_SC_OPEN_MAX);
     for (i = 0; i < open_max; i++)
     {
         /* don't hold open fd opened besides server socket */
-        if (i != fd)
+        if (i != server_fd)
             fcntl (i, F_SETFD, FD_CLOEXEC);
     }
 
@@ -821,11 +845,13 @@ int main(int argc, char** argv)
     g_io_channel_set_close_on_unref(ch, TRUE);
     g_io_add_watch(ch, G_IO_IN|G_IO_PRI, on_new_conn_incoming, NULL);
     g_io_add_watch(ch, G_IO_ERR, on_server_conn_close, NULL);
+    g_io_channel_unref(ch);
 
     g_type_init();
 
-    hash = g_hash_table_new_full(g_str_hash, g_str_equal,NULL,(GDestroyNotify)cache_unref);
+    hash = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, NULL);
 
+    main_loop = g_main_loop_new( NULL, TRUE );
     g_main_loop_run( main_loop );
     g_main_loop_unref( main_loop );
 
